#include "MorseCode.h"

// Morse code for ASCII characters 32-90 (SPACE to Z)
const uint8_t MorseCode::Morse[] PROGMEM = {
  0xF0, 0x2A, 0x44, 0x00, 0x00, 0x00, 0x00, 0x5B, 0x99, 0x99, 0x00, 0x00, 0xB5, 0x82, 0x49, 0x8A,
  0xCD, 0x5D, 0x2D, 0x1D, 0x16, 0x13, 0x83, 0xB3, 0xC3, 0xCA, 0xC1, 0xC1, 0x00, 0x86, 0x00, 0x28,
  0x54, 0x60, 0x87, 0x97, 0x80, 0x70, 0x27, 0xB0, 0x17, 0x30, 0x5E, 0x90, 0x47, 0xD0, 0xA0, 0xC0,
  0x57, 0xBE, 0x40, 0x10, 0xE0, 0x20, 0x1E, 0x50, 0x8E, 0x9E, 0xB7
};

char MorseCode::charBuffer[MORSE_BUFFER_SIZE];
uint16_t MorseCode::interval[16];
uint32_t MorseCode::prevMillis;
uint8_t MorseCode::writeIndex;
uint8_t MorseCode::readIndex;
uint8_t MorseCode::charIndex;
uint8_t MorseCode::dot_ms;
uint8_t MorseCode::pin;
bool MorseCode::idle;
bool MorseCode::done;

void MorseCode::initialize(uint8_t buzzerPin, float WPM) {
  pin = buzzerPin;
  pinMode(pin, OUTPUT);
  digitalWrite(pin, LOW);
  WPM = 1200.0 / WPM;
  if (WPM > 255.0) WPM = 255.0; // 255ms dots is about 4.7 words per minute
  else if (WPM < 10.0) WPM = 20.0; // 20ms dots is 60 words per minute
  dot_ms = (uint8_t)(WPM + 0.5);
  readIndex = writeIndex = 0;
  for (uint8_t i = 0; i < 16; i++)
    interval[i] = 0;
  charBuffer[writeIndex] = '\0';
  prevMillis = millis();
  idle = true;
}

void MorseCode::transmit(char *s) {
  if (!idle) return;
  for (uint8_t i = 0; i < MORSE_BUFFER_SIZE; i++) {
    if ((charBuffer[i] = *s++) == '\0')
      break;  
  }
  charIndex = 0;
  checkCharBuffer();
}

void MorseCode::transmit(const __FlashStringHelper *s) {
  if (!idle) return;
  uint8_t *p = (uint8_t PROGMEM *)s;
  for (uint8_t i = 0; i < MORSE_BUFFER_SIZE; i++) {
    if ((charBuffer[i] = pgm_read_byte_near(p + i)) == '\0')
      break;  
  }
  charIndex = 0;
  checkCharBuffer();
}
  
void MorseCode::loop(void) {
  uint32_t now = millis();
  uint16_t ms = (uint16_t)(now - prevMillis);
  prevMillis = now;
  uint16_t remainingInterval = interval[readIndex];
  if (remainingInterval) { // timing something
    if (remainingInterval == 0xFFFF) { // special SPACE marker interval
      remainingInterval = 0;
    }
    if (idle) { // was previously idle - so we should always be at an even index and we need to begin a sound
      idle = false;
      digitalWrite(pin, HIGH);
    }
    if (ms >= remainingInterval) { // time's up for this sound or interval
      ms -= remainingInterval; // any left-over milliseconds get used by the following interval
      interval[readIndex] = 0; // this interval is complete
      readIndex = (readIndex + 1) & 0x0F; // move to next interval
      if (readIndex == writeIndex) { // no more intervals
        digitalWrite(pin, LOW);
        idle = true;
      } else {
        digitalWrite(pin, readIndex & 0x01 ? LOW : HIGH); // start sounds when readIndex is even, finish them when readIndex is odd
        if ((writeIndex == readIndex + 1) || ((writeIndex == 0) && (readIndex == 15))) { // if this new interval is the last one
          checkCharBuffer(); // check for additional characters in buffer waiting to be transmitted
        }
      }
    }
    interval[readIndex] -= ms;
  }
}

void MorseCode::checkCharBuffer(void) {
  if (charIndex < MORSE_BUFFER_SIZE) {
    char c = charBuffer[charIndex];
    if (c) {
      if ((c >= 'a') && (c <= 'z')) {
        c = c - 'a' + 'A'; // toupper
      }
      charIndex++; 
      if ((c >= ' ') && (c <= 'Z')) {
        uint8_t code = pgm_read_byte_near(Morse + c - ' ');
        uint8_t nybble = (code >> 4) & 0x0F; // first look at most significant nybble of code
        done = false;
        while (nybble) {
          switch (nybble) {
            case 0x01:
              dot(); dot(); dot();
              break;
            case 0x02:
              dot(); dot(); dash();
              break;
            case 0x03:
              dot(); dot(); complete();
              break;
            case 0x04:
              dot(); dash(); dot();
              break;
            case 0x05:
              dot(); dash(); dash();
              break;
            case 0x06:
              dot(); dash(); complete();
              break;
            case 0x07:
              dot(); complete();
              break;
            case 0x08:
              dash(); dot(); dot();
              break;
            case 0x09:
              dash(); dot(); dash();
              break;
            case 0x0A:
              dash(); dot(); complete();
              break;
            case 0x0B:
              dash(); dash(); dot();
              break;
            case 0x0C:
              dash(); dash(); dash();
              break;
            case 0x0D:
              dash(); dash(); complete();
              break;
            case 0x0E:
              dash(); complete();
              break;
            case 0x0F: // space
              if (!idle) { // change existing inter-char interval to inter-word one (7 dot times instead of 3)
                space();
              }
              break;  
          }
          nybble = code & 0x0F; // now look at least significant nybble
          code = 0x00; // and after that we're done
          if (!nybble && !done) // if first nybble had three elements and second contains none, or if both nybbles contained three elements
            complete();
        }
      } 
    }
  }
}

void MorseCode::dot(void) {
  newInterval(dot_ms); // the dot itself
  newInterval(dot_ms); // the gap between the dot and the next sound
}

void MorseCode::dash(void) {
  newInterval(3 * dot_ms); // the dash itself
  newInterval(dot_ms); // the gap between the dash and the next sound
}

void MorseCode::space(void) { // a space
  newInterval(0xFFFF); // special marker interval produces a zero-length tone
  newInterval(dot_ms * 4); // there will already be a 3 dot length interval following the end of previous character - this expands it to 7 dot intervals which is the Morse standard for a space.
}

void MorseCode::newInterval(uint16_t ms) {
  interval[writeIndex] = ms;
  writeIndex = (writeIndex + 1) & 0x0F;
}

void MorseCode::complete(void) { // complete a Morse character by extending the final (silent) interval from a dot length to a dash length
  interval[writeIndex ? writeIndex - 1 : 0x0F] += 2 * dot_ms;
  done = true;
}

